/*
 * This file is part of Sponge, licensed under the MIT License (MIT).
 *
 * Copyright (c) SpongePowered <https://www.spongepowered.org>
 * Copyright (c) contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package org.spongepowered.common.registry.type.text;

import static org.spongepowered.common.text.SpongeTexts.COLOR_CHAR;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import org.spongepowered.api.registry.AdditionalCatalogRegistryModule;
import org.spongepowered.api.registry.CatalogRegistryModule;
import org.spongepowered.api.registry.RegistryModule;
import org.spongepowered.api.registry.util.RegisterCatalog;
import org.spongepowered.api.text.serializer.TextSerializer;
import org.spongepowered.api.text.serializer.TextSerializers;
import org.spongepowered.common.text.serializer.JsonTextSerializer;
import org.spongepowered.common.text.serializer.PlainTextSerializer;
import org.spongepowered.common.text.serializer.SpongeFormattingCodeTextSerializer;
import org.spongepowered.common.text.serializer.xml.TextXmlTextSerializer;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

public final class TextSerializerRegistryModule implements AdditionalCatalogRegistryModule<TextSerializer> {

    @RegisterCatalog(TextSerializers.class)
    private static final Map<String, TextSerializer> textSerializerMappings = new HashMap<>();

    @Override
    public Optional<TextSerializer> getById(String id) {
        return Optional.ofNullable(textSerializerMappings.get(id));
    }

    @Override
    public Collection<TextSerializer> getAll() {
        return ImmutableList.copyOf(textSerializerMappings.values());
    }

    @Override
    public void registerDefaults() {
        textSerializerMappings.put("plain", new PlainTextSerializer());
        textSerializerMappings.put("json", new JsonTextSerializer());
        textSerializerMappings.put("formatting_code", new SpongeFormattingCodeTextSerializer(
                "sponge:formatting_code", "Formatting Codes", '&'));
        textSerializerMappings.put("legacy_formatting_code", new SpongeFormattingCodeTextSerializer(
                "minecraft:legacy_formatting_code", "Legacy Formatting Codes", COLOR_CHAR));
        textSerializerMappings.put("text_xml", new TextXmlTextSerializer());
    }

    @Override
    public void registerAdditionalCatalog(TextSerializer serializer) {
        textSerializerMappings.put(serializer.getId(), serializer);
    }

}
